        PAGE 55, 132
        TITLE BASIC SASI HARD DISK DRIVER - REV 1.1
;------------------------------------------------------------------*
;  BASIC HARD DISK DRIVER FOR SASI.
;
;       Copyright 1985,86 by Turning Point Logic Corporation.
;       Permission granted for personal, non-commercial use.
;
;       This module is a basic hard disk driver for the SASI inter-
;       face.  It is designed to run on the IBM PCjr, IBM PC, or
;       compatable.
;
;       The hardware interface for the SASI interface is a very
;       simple option card that plugs into one of the open slots of
;       the PC.  The board mainly consists of buffers and is I/O
;       mapped on the PC Bus.  For hardware simplicity no DMA
;       is supported and the data transfers are performed entirely
;       by the CPU.  Since the 8086 has repeat instructions and
;       the driver is assumed to be running under DOS, which is a
;       single tasking system, throughput is comparable to those
;       hard disk devices utilizing the DMA hardware.
;
;       This driver will support up to eight devices and in the
;       near future will allow sharing of those devices by separate
;       host systems.
;
;       ENTRY:  SS:SP points to a control block:
;               SP+0  - Return address.
;               SP+2  - Caller type:
;                        0 - 'C' routine, only the least significant
;                            word of the data transfer address is
;                            valid.  The most significant word of
;                            the address is formed by using SS.
;                        1 - Device driver called, the data transfer
;                            address is a valid double word.
;               SP+4  - Command code:
;                        0 - Read sector.
;                        1 - Write sector.
;                        2 - Set Drive Characteristics.
;                        3 - Format drive.
;                        4 - Verify sector with data.
;                        5 - Test for drive ready.
;               SP+6  - Drive number (0-7).
;               SP+8  - Number of sectors to process (for commands
;                        0 & 1).
;               SP+10 - Starting logical sector number
;                        (used for commands 0 and 1, this is a double
;                         word with least significant byte first).
;               SP+14 - Data transfer area address (this is a double
;                        word in Intel format [segment and offset].
;               SP+18 - Interleave factor [1-16] (for command 3).
;               SP+20 - Number of cylinders (for command 2).
;               SP+22 - Number of heads (for command 2).
;               SP+24 - Starting cylinder number of reduced write
;                        current (for command 2).
;               SP+26 - Starting cylinder number for write precomp.
;               SP+28 - Maximum ECC data burst length (for
;                        command 2).
;               SP+30 - DCB control byte setting
;               SP+32 - Number of sector per track (word).
;       
;
;       EXIT:   AX contains a result code for all commands but Test
;                Drive Ready:
;                00.H - Operation successful.
;                01.H - No index detected from disk drive.
;                02.H - No seek complete from disk drive.
;                03.H - Write fault from disk drive.
;                04.H - Drive not ready after select.
;                06.H - Track 0 not found.
;                07.H - Data protected.
;                10.H - ID field read error.
;                11.H - Uncorrectable data error.
;                12.H - Address mark not found.
;                14.H - Target sector not found.
;                15.H - Seek error.
;                18.H - Correctable data error.
;                19.H - Bad track flag detected.
;                1A.H - Format error.
;                20.H - Invalid command.
;                21.H - Illegal disk address.
;                30.H - RAM diagnostic failure.
;                31.H - Program memory checksum error.
;                32.H - ECC diagnostic failure.
;                50.H - Media change.
;                81.H - Invalid controller number.
;                82.H - Timed out while selecting controller.
;                83.H - Bad controller status while issueing
;                        DCB.
;                84.H - Timed out while waiting for command
;                        to complete.
;                85.H - Bad controller status between drive
;                        status reads.
;                86.H - Drive status byte count exceeded.
;                87.H - Controller status not qualified.
;                88.H - Verify error.
;               AX contains a result code for Test Drive Ready
;                command:
;                0 - The media has not changed and the drive is ready
;                1 - The media has changed and the drive is ready
;                2 - An error occurred or the drive is not ready.
;
; ----- EQUATES.
;
D_MAX   EQU     8;      Maximum number of controllers.
R_CMD   EQU     8;      SASI read command code.
W_CMD   EQU     10;     SASI write command code.        
D_CMD   EQU     12;     SASI issue drive characteristic command.
F_CMD   EQU     4;      SASI format drive command.
T_CMD   EQU     0;      SASI Test drive ready command.
V_CMD   EQU     88H;    Internal verify sector command.
R_TRY   EQU     3;      Number of retries before aborting.
;
PROG    SEGMENT BYTE    PUBLIC  'PROG'
        ASSUME  CS:PROG,DS:NOTHING,SS:NOTHING,ES:NOTHING
PUBLIC  HDDRV
;
;
; ----- DATA CONTROL BLOCK (DCB).
;
CMMD    DB 0;           Command code.
CSECT   DB 3 DUP(0);    Logical sector number (msb first).
CCNT    DB 0;           Sector count.
CCTL    DB 0;           Control byte.
;
; ----- DRIVE CONTROL TABLE.
;
;       The Xebec controller for which this driver was designed,
;       has the capability of interfacing to two drives.  One 
;       limitation to this capability is that both drives must be
;       of the same size, type, from the same manufacturer, and
;       model number.  For this reason, the second drive is only
;       supported by this driver as a transparent extension of the
;       first drive.  In effect, from the caller's perspective the
;       Xebec controller has one large single drive rather than two
;       small drives.
;
;       To implement this feature, the number of sectors for either
;       drive is calculated (since the drives are of the same size
;       it does not matter which drive is used) and the result is
;       placed into a table.  One of the first items to be per-
;       formed before the drive (or controller actually) is put into
;       full use is to issue the drive characteristics to the Xebec
;       controller.  One of the entry parameters for the Issue Drive
;       Characteristic command is the SASI DCB Control Byte value.
;       This byte is the last byte of the six byte Data Control Block
;       and defines among other things half-step options and retry
;       parameters. At least one bit in this byte has not been defined
;       by Xebec.  Bit 5 is therefore used by this module as a flag
;       specifying whether one or two drives are connected to the
;       Xebec controller.
;
;       When this flag is set this driver will calculate the
;       maximum sector number for either drive, place that value
;       into the table below, and set the "two drive flag" in the
;       table for that particular controller.  On other commands
;       such as read sector or write sector, the two-drive-flag
;       is checked.  If it is set then the target sector number
;       is compared to the sector number in the table.  If the tar-
;       get sector is larger than the value in the table then the
;       sector number in the table is subtracted from the target
;       sector and the result is used as the new target sector but
;       on the second drive.
;
;       The table consists of D_MAX number of entries with 8 bytes
;       per entry.  Each entry defines one controller and is 
;       formated as shown below.
;
;
DCTBL   STRUC
;
DCTFLG  DB ?;                   Two-Drive-Flag.
DCTCV   DB ?;                   SASI DCB Control Byte value (with
;                                bit 5 cleared).
DCTSNL  DW ?;                   Negative number of sectors per 
DCTSNH  DW ?;                    drive (low word first).
DCTFIL  DW ?;                   Filler - unused (8 bytes total).
;
DCTBL   ENDS
;
;
DRCTBL  DB D_MAX*8 DUP(0);      Drive control table.
;
; ----- GENERAL DATA AREA.
;
RFLG    DB 0;           Sense drive status request in progress.
RSDCB   DB 6 DUP(0);    Request Sense Status DCB.
RSTAT   DB 4 DUP(?);    Sense drive status bytes.
CNUM    DB ?;           Controller number.
CNTRL   DB ?;           Controller number in bit mask form.
MMSG    DB ?;           Message byte from previous operation.
WCNT    DW ?;           Wait count remaining.
BCNT    DB ?;           Number of 512 byte blocks left to transfer.
RETRY   DB ?;           Retry counter.
VECNT   DW ?;           Verify error counter.
CBFLAG  DB ?;           Flag indicating that a second read, write,
;                        or verify operation is to be performed
;                        because the specified operation crossed
;                        the local drive boundry (two drives on
;                        one SASI controller).
NCCNT   DB ?;           Second operation transfer count in sectors.
;
; ----- DRIVE CHARACTERISTIC DATA BLOCK.
;
DCYLH   DB ?;           MSB of maximum number of cylinders.
DCYLL   DB ?;           LSB of maximum number of cylinders.
DHMAX   DB ?;           Maximum number of heads (1-4).
DRWCH   DB ?;           MSB of the starting reduced write current
;                        cylinder.
DRWCL   DB ?;           LSB of the starting reduced write current
;                        cylinder number.
DWPCH   DB ?;           MSB of the starting write precompensation
;                        cylinder number.
DWPCL   DB ?;           LSB of the starting write precompensation
;                        cylinder number.
DMECC   DB ?;           Maximum data burst length (usually
;                        set to 11).
;
; ----- MAIN ENTRY POINT.
;
HDDRV   PROC    NEAR
;
        PUSH BP;                        Set pointer to variables.
        MOV BP, SP;
        ADD BP, 2;
;
        PUSH BX;                        Save work registers.
        PUSH CX;
        PUSH DX;
        PUSH DI;
        PUSH SI;
        PUSH DS;
        PUSH ES;
;
        CMP WORD PTR [BP]+6, D_MAX;     Check controller number
        MOV AX, 81H;                     for validity.
        JL SLCM;
        JMP SRTN;
;
SLCM:   MOV AX, [BP]+4;                 Load and check command code
        CMP AX, 2;                       for format or device
        JL STRW;                         status commands.
        JE SDCD;
        CMP AX, 3;
        JE SDC1;
        CMP AX, 4;
        JE SDC2;
        CMP AX, 5;
        JE SDC3;
        MOV AX, 88H;
        JMP SRTN;
SDCD:   JMP SIDC;
SDC1:   JMP SFMT;
SDC2:   MOV BL, V_CMD;
        JMP STCMD;
SDC3:   JMP STDR;
;
; ----- HANDLE SECTOR READ/WRITE/VERIFY COMMANDS.
;
STRW:   MOV CBFLAG, 0;                  Clear split-operation flag.
;
        MOV BL, R_CMD;                  Determine and set the SASI
        CMP AX, 0;                       command code.
        JE STCMD;
        MOV BL, W_CMD;
STCMD:  MOV CMMD, BL;
;
        MOV AX, [BP]+8;                 Load and store the sector
        MOV CCNT, AL;                    count.
;
        MOV DX, [BP]+6;                 Load controller number.
        MOV AX, DX;
;
        MOV DI, OFFSET DRCTBL;          Index into the Drive Control
        SHL AX, 1;                       Table via the controller 
        SHL AX, 1;                       number.
        SHL AX, 1;
        ADD DI, AX;
;
        MOV AL, CS:[DI].DCTCV;          Set the SASI DCD Control
        MOV CCTL, AL;                    byte from the value found
;                                        in the table.
;
        CMP CS:[DI].DCTFLG, 0;          Test for the Two-Drive
        JE STLSN;                        Flag set.  If is not
;                                        then simply load the
;                                        the target sector number.
;
        MOV AX, CS:[DI].DCTSNL;         Otherwise, add the target
        ADD AX, [BP]+10;                 sector number to the nega-
        MOV CX, AX;                      tive sector number in the
        MOV AX, CS:[DI].DCTSNH;          table.  If the result is
        ADC AX, [BP]+12;                 still negative then the
        JNC STCDB;                       target sector is on the
        OR AL, 00100000B;                the first drive.  If the   
        MOV AL, CSECT;                   result is positive then    
        MOV CH, CSECT+1;                 the target sector is the   
        MOV CL, CSECT+2;                 result and is on the       
        JMP STLDTA;                      second drive.              
;
STCDB:  PUSH DX;                        Add to the resulting nega-
        MOV DX, AX;                      tive value (negative number
        MOV AL, CCNT;                    of sectors to end of first
        XOR AH, AH;                      drive) with the transfer
        ADD CX, AX;                      count.  If the result is
        ADC DX, 0;                       again negative or zero, 
        POP DX;                          then the entire operation
        JNC STLSN;                       is to be performed on the
        CMP CX, 0;                       first drive.
        JE STLSN;
;
        MOV NCCNT, CL;                  Save the sector count to be
        MOV AL, CCNT;                    performed on the second
        SUB AL, CL;                      drive and calculate the 
        MOV CCNT, AL;                    sector count for the first
        MOV CBFLAG, 1;                   drive.  Set control flag
;                                        indicating that two opera-
;                                        tions are to be performed.
;
STLSN:  MOV AX, [BP]+10;                Copy the starting logical
        MOV CSECT+2, AL;                 sector number into the DCB.
        MOV CSECT+1, AH;
        MOV AL, [BP]+12;
        MOV CSECT, AL;
;
STLDTA: CMP WORD PTR [BP]+2, 0;         Load data transfer address.
        JNE STLAD;                       [set segment value of
        MOV AX, DS;                       address on whether the
        MOV ES, AX;                       caller was a 'C' routine
        MOV DI, [BP]+14;                  or the SASI driver].
        JMP SHORT STDCB;
STLAD:  LES DI, [BP]+14;                
;
STDCB:  MOV SI, OFFSET CMMD;            Set pointer to DCB.
;
STRWL:  CALL ISCMD;                     Perform read or write opera-
        CMP AX, 0;                       tion.  If unsuccessful,
        JE STNXT;                        then return immediately.
        JMP SRTN;
;
STNXT:  CMP CBFLAG, 0;                  Check for a second operation
        JNE STNXT1;                      to be performed.
        JMP SRTN;
;
STNXT1: MOV AL, CCNT;                   If so then address the data
        XOR AH, AH;                      transfer address for the 
        ADD DI, AX;                      second operation and
        MOV CSECT+2, 0;                  set sector zero,                
        MOV CSECT+1, 0;
        MOV CSECT, 00100000B;            set second drive on cntlr,
        MOV AL, NCCNT;                   remaining number of sectors,
        MOV CCNT, AL;
        MOV DX, [BP]+6;                  and restore controller #.
;
        CALL ISCMD;                     Perform second read, write,
        JMP SRTN;                        verify operation on the
;                                        remaining number of sectors.
;
;
; ----- HANDLE DRIVE CHARACTERISTIC COMMAND.
;
SIDC:   MOV CMMD, D_CMD;                Set issue drive char-
;                                        acteristic command in DCB.
;
        MOV AX, [BP]+22;                Load and store the maximum
        AND AL, 0FH;                     number of heads.
        MOV DHMAX, AL;
;
        MOV AX, [BP]+24;                Load and store the starting
        MOV DRWCH, AH;                   cylinder number of reduced
        MOV DRWCL, AL;                   write current.
;
        MOV AX, [BP]+26;                Load and store the starting
        MOV DWPCH, AH;                   cylinder number of write
        MOV DWPCL, AL;                   precompensation.
;
        MOV AX, [BP]+28;                Load and store the maximum
        AND AL, 0FH;                     number of ECC data burst
        MOV DMECC, AL;                   length.
;
        MOV DX, [BP]+6;                 Index into the Drive Control
        MOV AX, DX;                      table via the controller
        SHL AX, 1;                       number.
        SHL AX, 1;
        SHL AX, 1;
        MOV DI, OFFSET DRCTBL;
        ADD DI, AX;
;
        MOV AX, [BP]+30;                Load and store the SASI DCB
        AND AL, 11011111B;               Control byte into the
        MOV CS:[DI].DCTCV, AL;           table.
;
        MOV CS:[DI].DCTFLG, 0;          Check for the two-drive
        MOV AX, [BP]+30;                 flag bit set in the
        AND AL, 00100000B;               SASI Control Byte.
        JZ ISCMX;
;
        MOV CS:[DI].DCTFLG, 1;          If it is then set the two-
        MOV DX, [BP]+20;                 drive-flag in the table,
        SHR DX, 1;                       calculate the number of
        MOV AX, [BP]+32;                 sectors per drive by
        MUL DX;                          multiplying the number of
        MOV BX, AX;                      cylinders times the number
        MOV CX, DX;                      of sectors per track times
        MOV SI, [BP]+22;                 the number of heads.
IDSML:  ADD AX, BX;                     [Note: due to the lack of
        ADC DX, CX;                      32-bit arithmetic in the
        DEC SI;                          8088, the last multiply
        JNZ IDSML;                       is performed via successive
;                                        addition].
;
        NOT AX;                         Negate the resulting value
        NOT DX;                          and place result into the
        ADD AX, 1;                       table.
        ADC DX, 0;
        MOV CS:[DI].DCTSNL, AX;
        MOV CS:[DI].DCTSNH, DX;
;
        MOV DX, [BP]+6;                 Load the controller number,
        MOV AX, [BP]+20;                 and the number of cylinders
        SHR AX, 1;                       divided by two.
        JMP SHORT ISCMX1;
;
ISCMX:  MOV AX, [BP]+20;                Load and store the maximum
ISCMX1: MOV DCYLH, AH;                   number of cylinders count.
        MOV DCYLL, AL;
;
        MOV AX, CS;                     Set pointer to drive char-
        MOV ES, AX;                      acteristics data.
        MOV DI, OFFSET DCYLH;
;
        MOV SI, OFFSET CMMD;            Set pointer to DCB.
;
        MOV DL, [BP]+6;                 Load controller number.
;
        CALL ISCMD;                     Issue command and exit.
        JMP SRTN;
;
; ----- HANDLE FORMAT COMMAND.
;
SFMT:   MOV CMMD, F_CMD;                Set format command in DCB.
;
        MOV AL, 0;                      Clear the starting track
        MOV CSECT, AL;                   number.
        MOV CSECT+1, AL;
        MOV CSECT+2, AL;
;
        MOV DX, [BP]+6;                 Load controller number.
        MOV AX, DX;
;
        MOV DI, OFFSET DRCTBL;          Index into the Drive Control
        SHL AX, 1;                       Table via the controller 
        SHL AX, 1;                       number.
        SHL AX, 1;
        ADD DI, AX;
;
        MOV AL, CS:[DI].DCTCV;          Set the SASI DCD Control
        MOV CCTL, AL;                    byte from the value found
;                                        in the table.
;
        MOV AX, [BP]+18;                Load and set the interleave
        AND AL, 0FH;                     factor.
        MOV CCNT, AL;
;
        MOV SI, OFFSET CMMD;            Set pointer to DCB.
;
        CALL ISCMD;                     Perform the format operation.
;
        JMP SRTN;                       Exit.
;
; ----- HANDLE TEST DRIVE READY COMMAND.
;
STDR:   MOV CMMD, T_CMD;                Set Test Drive Ready command
;                                        code.
;
        MOV DX, [BP]+6;                 Load controller number.
        MOV AX, DX;
;
        MOV DI, OFFSET DRCTBL;          Index into the Drive Control
        SHL AX, 1;                       Table via the controller 
        SHL AX, 1;                       number.
        SHL AX, 1;
        ADD DI, AX;
;
        MOV AL, CS:[DI].DCTCV;          Set the SASI DCD Control
        MOV CCTL, AL;                    byte from the value found
;                                        in the table.
;
        MOV SI, OFFSET CMMD;            Set pointer to DCB.
;
        CALL ISCMD;                     Perform the operation.
;
        CMP AX, 0;                      If the operation was
        JE SRTN;                         successful then the drive
        CMP AX, 4;                       is ready and the media
        JNE STDM;                        has not changed.  Other-
        MOV AX, 1;                       wise test for drive not
        JMP SHORT SRTN;                  ready error or media
STDM:   CMP AX, 50H;                     change result.  If so, then
        JNE STDE;                        exit with media change
        MOV AX, 1;                       result code.  Set general
        JMP SHORT SRTN;                  error failure for any other
STDE:   MOV AX, 2;                       type of error.
;
; ----- RETURN.
;
SRTN:   POP ES;                         Restore work registers and
        POP DS;                          return.
        POP SI;
        POP DI;
        POP DX;
        POP CX;
        POP BX;
        POP BP;
        RET;
;
PAGE 
;------------------------------------------------------------------*
;  ISSUE COMMANDS TO SASI DEVICE.
;
;       This routine is the lowest level module of the SASI device
;       driver.  It receives on entry Data Control Blocks (DCB)
;       which are to be transferred over the SASI bus to the target
;       device.  After transferring the DCB it will either recieve
;       or send the corresponding data (e.g. sector data).  After
;       which it will retrieve the completion message and interogate
;       if there is an error.
;
;       ENTRY:  DL contains the controller number (0-7).
;               CS:SI points to the DCB.
;               ES:DI points to the data transfer area (if applic-
;                able).
;
;       EXIT:   AX contains a result code:
;                00.H - Operation successful.
;                01.H - No index detected from disk drive.
;                02.H - No seek complete from disk drive.
;                03.H - Write fault from disk drive.
;                04.H - Drive not ready after select.
;                06.H - Track 0 not found.
;                07.H - Data protected.
;                10.H - ID field read error.
;                11.H - Uncorrectable data error.
;                12.H - Address mark not found.
;                14.H - Target sector not found.
;                15.H - Seek error.
;                18.H - Correctable data error.
;                19.H - Bad track flag detected.
;                1A.H - Format error.
;                20.H - Invalid command.
;                21.H - Illegal disk address.
;                30.H - RAM diagnostic failure.
;                31.H - Program memory checksum error.
;                32.H - ECC diagnostic failure.
;                50.H - Media change.
;                81.H - Invalid controller number.
;                82.H - Timed out while selecting controller.
;                83.H - Bad controller status while issueing DCB.
;                84.H - Timed out while waiting for command to com-
;                        plete.
;                85.H - Bad controller status between drive
;                        status reads.
;                86.H - Drive status byte count exceeded.
;                87.H - Controller status not qualified.
;                88.H - Verify error.
;
;               The registers BX, CX, and DX are destroyed.
;
; ----- EQUATES.
;
T_CNT           EQU 60600;      Wait count (approximately 4 seconds).
;
REQ_STS         EQU 3;          Request sense status command code.
;
SCT_SIZE        EQU 512;        Hard disk sector size in bytes.
;
SASI_RST        EQU 3FFAH;      SASI Reset port (LSB).
SASI_SEL        EQU 3FFBH;      SASI Select port.
SASI_WDT        EQU 3FFCH;      SASI Write Data port.
SASI_RDT        EQU 3FFDH;      SASI Read Data port.
SASI_STS        EQU 3FFEH;      SASI Status port.
;
;
ISCMD:  PUSH ES;                        Save work registers.
        PUSH DS;
        PUSH SI;
        PUSH DI;
;
        MOV CNUM, DL;                   Save the controller number.

        MOV CL, DL;                     Convert the binary
        MOV AH, 1;                       controller number into
        SHL AH, CL;                      a bit mask.
        MOV CNTRL, AH;
;
; ----- CHECK FOR CONTROLLER AVAILABLE.
;
        MOV CX, T_CNT;                  Set timeout counter
        MOV DX, SASI_STS;                and retry flag.
;
IRLP:   IN AL, DX;                      Continue checking for the
        AND AL, 1FH;                     controller to be avail-
        JZ ISEL;                         able.
        CALL SWT;
        DEC CX;
        JNZ IRLP;
;
        MOV AX, 82H;                    If the controller was not       
        JMP ISRTN;                       available after a period       
;                                        time then exit with error.      
;
; ----- SELECT CONTROLLER.
;
ISEL:   MOV AL, CNTRL;                  Attempt to select the
        NOT AL;                          controller.
        MOV DX, SASI_WDT;
        OUT DX, AL;
        MOV DX, SASI_SEL;
        OUT DX, AL;
;
        MOV CL, 0;                      Wait for the controller to      
        MOV DX, SASI_STS;                assert the request bit,               
ISELP:  IN AL, DX;                       first in a tight loop          
        TEST AL, 1;                      and if it is not asserted      
        JNZ ICST;                        then loop checking while       
        DEC CL;                          pausing in between.
        JNZ ISELP;                      
        MOV CX, T_CNT;
ISELP1: IN AL, DX;
        TEST AL, 1;
        JNZ ICST;
        CALL SWT;
        DEC CX;
        JNZ ISELP1;
        MOV AX, 87H;
        JMP ISRTN;
;
ICST:   AND AL, 1FH;                    After the request bit is set
        CMP AL, 00010011B;               check for a controller
        JE ITDCB;                        ready status.  If not then
        MOV AX, 83H;                     exit with error.
        JMP ISRTN;
;
; ----- TRANSFER DCB.
;
ITDCB:  MOV AX, CS;                     Set DS point to the segment
        MOV DS, AX;                      value of the DCB.
;
        MOV AL, CS:[SI];                Load and save the DCB
        PUSH AX;                         code and strip off the
        AND BYTE PTR CS:[SI], 7FH;       the least significant bit.
;
        MOV BX, SI;                     Set loop parameters for
        MOV CX, 6;                       transferring DCB.
;
IDCB:   MOV DX, SASI_STS;               Check for the DMA request   
        IN AL, DX;                       bit set.  Wait if necessary
        TEST AL, 1;                      and exit if timeout occurs.
        JNZ ITRN;                       
        CALL IDWT;
        JC ITRN;
        POP AX;
        MOV CS:[SI], AL;
        MOV AX, 83H;
        JMP ISRTN;
;
ITRN:   MOV AL, [BX];                   Transfer one byte of the DCB
        MOV DX, SASI_WDT;                and loop for all bytes.
        NOT AL;
        OUT DX, AL;
        INC BX;
        LOOPNZ IDCB;
;
        POP AX;                         Restore the SASI DCD command
        MOV CS:[SI], AL;                 code.
;
        MOV WCNT, T_CNT;                Set timeout count.
;
        CMP AL, 3;                      Determine the transfer          
        JNE ITYPE;                       type and direction [e.g.       
        JMP IRRD;                        read vs write sector].         
ITYPE:  CMP AL, 8;
        JNE ITYPE1;
        JMP IRRD;
ITYPE1: CMP AL, 10;
        JE IWWT;
        CMP AL, 12;
        JE IWWT;
        CMP AL, 88H;
        JNE ITGN;
        JMP IVER;
ITGN:   JMP ISTAT;
;
PAGE
; ----- TRANSFER DATA BYTES TO THE SASI INTERFACE.
;
IWWT:   MOV CL, 0;                      Before starting the transfer    
        MOV DX, SASI_STS;                make sure the controller is           
IWWT1:  IN AL, DX;                       ready for the transfer by      
        AND AL, 1FH;                     first checking in a tight      
        CMP AL, 00010001B;               loop.
        JE ISWRT;                       
        DEC CL;
        JNZ IWWT1;
;
IWWT2:  IN AL, DX;                      If the controller is still
        AND AL, 1FH;                     not ready for the transfer
        CMP AL, 00010001B;               then continue checking but
        JE ISWRT;                        pause briefly in between
        CALL SWT;                        until a certain length of
        DEC WCNT;                        time has expired.
        JNZ IWWT2;
        JMP ISTAT;
;
ISWRT:  MOV AX, ES;                     Set DS:BX with the pointer  
        MOV DS, AX;                      of the write data.
        MOV BX, DI;
;
        CMP BYTE PTR CS:[SI], 12;       Check for Initialize Drive      
        JNE IWTRN;                       Characteristics command.       
        MOV CX, SASI_STS;                If it is then only transfer    
        MOV DX, SASI_WDT;                eight bytes.
        MOV CH, -1;                     
        MOV AH, 1;                      
        JMP IWRT8;
;
IWTRN:  MOV CH, CS:[SI]+4;              Set loop count.
;
IWSSZ:  MOV AH, SCT_SIZE/16;            Set sector size loop count.
        MOV DX, SASI_WDT;
;
IWRT:   MOV AL, [BX];                   Transfer 16 bytes of data.
        NOT AL;
        OUT DX, AL;
        INC BX;
        MOV AL, [BX];                   NOTE:  This redundant code      
        NOT AL;
        OUT DX, AL;                     is added to increase            
        INC BX;                         throughput.  ANY alteration     
        MOV AL, [BX];                   of this block WILL greatly      
        NOT AL;
        OUT DX, AL;                     reduce the throughput of        
        INC BX;                         of the SASI driver.             
        MOV AL, [BX];
        NOT AL;
        OUT DX, AL;
        INC BX;
        MOV AL, [BX];
        NOT AL;
        OUT DX, AL;
        INC BX;
        MOV AL, [BX];
        NOT AL;
        OUT DX, AL;
        INC BX;
        MOV AL, [BX];
        NOT AL;
        OUT DX, AL;
        INC BX;
        MOV AL, [BX];
        NOT AL;
        OUT DX, AL;
        INC BX;
IWRT8:  MOV AL, [BX];
        NOT AL;
        OUT DX, AL;
        INC BX;
        MOV AL, [BX];
        NOT AL;
        OUT DX, AL;
        INC BX;
        MOV AL, [BX];
        NOT AL;
        OUT DX, AL;
        INC BX;
        MOV AL, [BX];
        NOT AL;
        OUT DX, AL;
        INC BX;
        MOV AL, [BX];
        NOT AL;
        OUT DX, AL;
        INC BX;
        MOV AL, [BX];
        NOT AL;
        OUT DX, AL;
        INC BX;
        MOV AL, [BX];
        NOT AL;
        OUT DX, AL;
        INC BX;
        MOV AL, [BX];
        NOT AL;
        OUT DX, AL;
        INC BX;
;
IWRTL:  DEC AH;                         Decrement the loop count and
        JZ IWRTL1;                       loop for all blocks.  When
        JMP IWRT;                        complete retrieve the  
IWRTL1: CMP CH, -1;                      drive status.
        JNE IWCLP;                       
IWJMP:  JMP ISTAT;
IWCLP:  DEC CH;
        JZ IWJMP;
;
        MOV CL, 0;                      Before looping to the next    
        MOV DX, SASI_STS;                sector make sure the    
IWNX1:  IN AL, DX;                       controller is ready by      
        AND AL, 1FH;                     first checking in a tight      
        CMP AL, 00010001B;               loop.
        JE IWNXL;                       
        DEC CL;
        JNZ IWNX1;
;
        MOV WCNT, T_CNT;
IWNX2:  IN AL, DX;                      If the controller is still
        AND AL, 1FH;                     not ready for the transfer
        CMP AL, 00010001B;               then continue checking but
        JE IWNXL;                        pause briefly in between
        CALL SWT;                        until a certain length of
        DEC WCNT;                        time has expired.
        JNZ IWNX2;
        JMP ISTAT;
;
IWNXL:  JMP IWSSZ;                      Loop for next block (sector).
;
PAGE
; ----- RECEIVE DATA BYTES FROM THE SASI INTERFACE.
;
IRRD:   MOV CL, 0;                      Before starting the transfer    
        MOV DX, SASI_STS;                make sure the controller is    
IRRD1:  IN AL, DX;                       ready for the transfer by      
        AND AL, 1FH;                     first checking in a tight      
        CMP AL, 00010101B;               loop.
        JE ISRD;                        
        DEC CL;
        JNZ IRRD1;
;
IRRD2:  IN AL, DX;                      If the controller is still
        AND AL, 1FH;                     not ready for the transfer
        CMP AL, 00010101B;               then continue checking but
        JE ISRD;                         pause briefly in between
        CALL SWT;                        until a certain length of
        DEC WCNT;                        time has expired.
        JNZ IRRD2;
        JMP ISTAT;
;
ISRD:   MOV AX, ES;                     Set DS:BX with the pointer  
        MOV DS, AX;                      for the read data.
        MOV BX, DI;
;
        CMP BYTE PTR CS:[SI], 3;        Check for Request Sense         
        JNE IRTRN;                       Status command.  If it is      
        MOV CX, SASI_STS;                then only transfer four        
        MOV DX, SASI_RDT;                bytes.
        MOV CH, -1;                     
        MOV AH, 1;                      
        JMP IRD4;
;
IRTRN:  MOV CH, CS:[SI]+4;              Set loop count.
;
IRSSZ:  MOV AH, SCT_SIZE/16;            Set sector size loop count.
        MOV DX, SASI_RDT;
;
IRD:    IN AL, DX;                      Transfer 16 bytes of data.
        MOV [BX], AL;
        INC BX;
        IN AL, DX;                      NOTE:  This redundant code      
        MOV [BX], AL;                    is added to increase            
        INC BX;                          throughput.  ANY alteration     
        IN AL, DX;                       of this block WILL greatly      
        MOV [BX], AL;                    reduce the throughput of        
        INC BX;                          of the SASI driver.
        IN AL, DX;
        MOV [BX], AL;
        INC BX;
        IN AL, DX;
        MOV [BX], AL;
        INC BX;
        IN AL, DX;
        MOV [BX], AL;
        INC BX;
        IN AL, DX;
        MOV [BX], AL;
        INC BX;
        IN AL, DX;
        MOV [BX], AL;
        INC BX;
        IN AL, DX;
        MOV [BX], AL;
        INC BX;
        IN AL, DX;
        MOV [BX], AL;
        INC BX;
        IN AL, DX;
        MOV [BX], AL;
        INC BX;
        IN AL, DX;
        MOV [BX], AL;
        INC BX;
IRD4:   IN AL, DX;
        MOV [BX], AL;
        INC BX;
        IN AL, DX;
        MOV [BX], AL;
        INC BX;
        IN AL, DX;
        MOV [BX], AL;
        INC BX;
        IN AL, DX;
        MOV [BX], AL;
        INC BX;
;
IRDL:   DEC AH;                         Decrement the loop count and
        JZ IRDL1;                        loop for all blocks.  When     
        JMP IRD;                         complete retrieve the          
IRDL1:  CMP CH, -1;                      drive status.
        JE IRDL2;                        
        DEC CH;
        JNZ IRNXS;
IRDL2:  JMP ISTAT;
;
IRNXS:  MOV CL, 0;                      Before looping to the next    
        MOV DX, SASI_STS;                sector make sure the    
IRNX1:  IN AL, DX;                       controller is ready by      
        AND AL, 1FH;                     first checking in a tight      
        CMP AL, 00010101B;               loop.
        JE IRNXL;                       
        DEC CL;
        JNZ IRNX1;
;
        MOV WCNT, T_CNT;
IRNX2:  IN AL, DX;                      If the controller is still
        AND AL, 1FH;                     not ready for the transfer
        CMP AL, 00010101B;               then continue checking but
        JE IRNXL;                        pause briefly in between
        CALL SWT;                        until a certain length of
        DEC WCNT;                        time has expired.
        JNZ IRNX2;
        JMP ISTAT;
;
IRNXL:  JMP IRSSZ;
;
PAGE    
; ----- VERIFY DATA BYTES FROM THE SASI INTERFACE.
;
IVER:   MOV VECNT, 0;                   Clear the verify error count.
;
        MOV CL, 0;                      Before starting the transfer    
        MOV DX, SASI_STS;                make sure the controller is    
IVER1:  IN AL, DX;                       ready for the transfer by      
        AND AL, 1FH;                     first checking in a tight      
        CMP AL, 00010101B;               loop.
        JE ISVR;                        
        DEC CL;
        JNZ IVER1;
;
IVER2:  IN AL, DX;                      If the controller is still
        AND AL, 1FH;                     not ready for the transfer
        CMP AL, 00010101B;               then continue checking but
        JE ISVR;                         pause briefly in between
        CALL SWT;                        until a certain length of
        DEC WCNT;                        time has expired.
        JNZ IVER2;
        JMP ISTAT;
;
ISVR:   MOV AX, ES;                     Set DS:BX with the pointer  
        MOV DS, AX;                      for the read data.
        MOV BX, DI;
;
        MOV CH, CS:[SI]+4;              Set loop count.
;
IVSSZ:  MOV AH, SCT_SIZE/16;            Set sector size loop count.
        MOV DX, SASI_RDT;
;
IVR:    IN AL, DX;                      Verify 16 bytes of data.
        CMP [BX], AL;       
        JE IVI;
        CALL IVERR;
IVI:    INC BX;                 
        IN AL, DX;                      NOTE:  This redundant code      
        CMP [BX], AL;                    is added to increase            
        JE IVI1;                         throughput.  ANY alteration     
        CALL IVERR;                      of this block WILL greatly      
IVI1:   INC BX;                          reduce the throughput of        
        IN AL, DX;                       of the SASI driver.             
        CMP [BX], AL;
        JE IVI2;
        CALL IVERR;
IVI2:   INC BX;
        IN AL, DX;
        CMP [BX], AL;
        JE IVI3;
        CALL IVERR;
IVI3:   INC BX;
        IN AL, DX;
        CMP [BX], AL;
        JE IVI4;
        CALL IVERR;
IVI4:   INC BX;
        IN AL, DX;
        CMP [BX], AL;
        JE IVI5;
        CALL IVERR;
IVI5:   INC BX;
        IN AL, DX;
        CMP [BX], AL;
        JE IVI6;
        CALL IVERR;
IVI6:   INC BX;
        IN AL, DX;
        CMP [BX], AL;
        JE IVI7;
        CALL IVERR;
IVI7:   INC BX;
        IN AL, DX;
        CMP [BX], AL;
        JE IVI8;
        CALL IVERR;
IVI8:   INC BX;
        IN AL, DX;
        CMP [BX], AL;
        JE IVI9;
        CALL IVERR;
IVI9:   INC BX;
        IN AL, DX;
        CMP [BX], AL;
        JE IVI10;
        CALL IVERR;
IVI10:  INC BX;
        IN AL, DX;
        CMP [BX], AL;
        JE IVI11;
        CALL IVERR;
IVI11:  INC BX;
        IN AL, DX;
        CMP [BX], AL;
        JE IVI12;
        CALL IVERR;
IVI12:  INC BX;
        IN AL, DX;
        CMP [BX], AL;
        JE IVI13;
        CALL IVERR;
IVI13:  INC BX;
        IN AL, DX;
        CMP [BX], AL;
        JE IVI14;
        CALL IVERR;
IVI14:  INC BX;
        IN AL, DX;
        CMP [BX], AL;
        JE IVI15;
        CALL IVERR;
IVI15:  INC BX;
;
IVRL:   DEC AH;                         Decrement the loop count and
        JZ IVRL1;                        loop for all blocks.  When
        JMP IVR;                         complete retrieve the
IVRL1:  CMP CH, -1;                      drive status.
        JE ISTAT;
        DEC CH;
        JZ ISTAT;
;
        MOV CL, 0;                      Before looping to the next    
        MOV DX, SASI_STS;                sector make sure the    
IVNX1:  IN AL, DX;                       controller is ready by      
        AND AL, 1FH;                     first checking in a tight      
        CMP AL, 00010101B;               loop.
        JE IVNXL;                       
        DEC CL;
        JNZ IVNX1;
;
        MOV WCNT, T_CNT;
IVNX2:  IN AL, DX;                      If the controller is still
        AND AL, 1FH;                     not ready for the transfer
        CMP AL, 00010101B;               then continue checking but
        JE IVNXL;                        pause briefly in between
        CALL SWT;                        until a certain length of
        DEC WCNT;                        time has expired.
        JNZ IVNX2;
        JMP ISTAT;
;
IVNXL:  JMP IVSSZ;                      Loop for next block (sector).
;
;
IVERR:  INC VECNT;                      Increment the verify error
        RET;                             count.
; 
; ----- RETRIEVE DRIVE STATUS.
;
ISTAT:  MOV BX, T_CNT;                  Set timeout counter.
        MOV CL, 0;
        MOV AL, CS:[SI];
        CMP AL, 4;
        JNE ISTST;
        MOV BX, 23470;                  [5 minutes * 60/66 us]
        MOV CL, 69;
;
ISTST:  MOV CH, 0;                      Wait in a tight loop at         
        MOV DX, SASI_STS;                first for the controller       
ISWT:   IN AL, DX;                       to return the command          
        AND AL, 1FH;                     completion status.
        CMP AL, 00010111B;              
        JE IRCST;
        DEC CH;
        JNZ ISWT;
;
ISWT1:  IN AL, DX;                      If the controller is still
        AND AL, 1FH;                     not ready, still continue
        CMP AL, 00010111B;               checking but pause
        JE IRCST;                        briefly in between.
        CALL SWT;
        DEC BX;
        JNZ ISWT1;
        DEC CL;
        JNS ISWT1;
        MOV AX, 84H;
        JMP ISRTN;
;
IRCST:  MOV DX, SASI_RDT;               Retrieve the two command
        IN AL, DX;                       completion status bytes.
        MOV AH, AL;
        CALL IDWT;
        JC IRSC1;
        MOV AX, 86H;
        JMP ISRTN;
IRSC1:  IN AL, DX;
        MOV MMSG, AL;
;
        CMP AH, 0;                      Check for successful com-
        JNE IRDST;                       pletion.  Also, check
        MOV AX, 0;                       for verify count not zero
        MOV BL, CS:[SI];                 if this was a verify
        CMP BL, 88H;                     operation.  Exit with
        JNE ISRTN;                       the appropriate result
        CMP VECNT, 0;                    code.
        JE ISRTN;
        MOV AX, 88H;
        JMP ISRTN;
;
; ----- REQUEST DRIVE STATUS ON ERROR.
;
IRDST:  CMP RFLG, 0;                    Determine if this request
        JE IRDSP;                        is a drive status request.
        MOV AX, 85H;                     If it is and an error 
        JMP ISRTN;                       occurred, then exit
;                                        immediately.
;
IRDSP:  MOV RFLG, 1;                    Prepare the sense drive
        MOV SI, OFFSET RSDCB;            status DCB.
        MOV BYTE PTR CS:[SI], REQ_STS;
;
        MOV AX, CS;                     Set the data address to
        MOV ES, AX;                      store the drive status
        MOV DI, OFFSET RSTAT;            bytes and load the
        MOV DL, CNUM;                    controller number.
;
        CALL ISCMD;                     Determine if the sense
        CMP AX, 0;                       drive status operation
        JNE ISRTN;                       was successful.  If not,
;                                        exit immediately.
;
        POP DI;                         Restore the data pointer
        POP SI;                          and the DCB pointer.
        POP DS;
        POP ES;
        PUSH ES;
        PUSH DS;
        PUSH SI;
        PUSH DI;
;
        XOR AH, AH;                     Retrieve and adjust the
        MOV AL, RSTAT;                   result.
        AND AL, 127;
        JNZ ISRTN;
        CMP BYTE PTR CS:[SI], 0;
        JNE ISRTN;
        MOV AX, 80H;
;
; ----- RETURN.
;
ISRTN:  POP DI;                         Restore work registers,
        POP SI;                          clear the drive status
        POP DS;                          request flag and return.
        POP ES;
        MOV RFLG, 0;
        RET;
;
;
; ----- RESET THE SASI BUS.
;
;       This routine will reset all of the controllers attached to
;       the SASI bus.
;
;       ENTRY: No specific entry conditions.
;
;       EXIT:  The SASI bus has been issued a reset.
;
;
PUBLIC  IRSASI
;
IRSASI: PUSH AX;                        Save work register.
        PUSH DX;
;
        CLI;                            Disable interrupts.
;
        MOV DX, SASI_WDT;               Clear the SEL line by
        MOV AL, 0FFH;                    writing a null byte to the
        OUT DX, AL;                      Write-data port.
;
        MOV AL, 1;                      Set reset bit and loop count.
        MOV AH, 10;
;
        MOV DX, SASI_RST;               Reset the SASI bus.
        OUT DX, AL;
;
RWLP:   DEC AH;                         Pause for approximately
        JNZ RWLP;                        28 microseconds.
;
        MOV AL, 0;                      Release SASI bus from reset.
        OUT DX, AL;     
;
        STI;                            Reenable interrupts.
;
        POP DX;                         Restore work registers and
        POP AX;                          return.
        RET;
;
;
; ----- WAIT SUBROUTINE.
;
;       ENTRY: No specific entry conditions.
;
;       EXIT:  Approximately 66 microseconds has expired.
;
;
SWT:    PUSH AX;                        Save work register.
;
        MOV AX, 9;                      Set loop count.
;
SWTLP:  DEC AX;                         Loop until loop count
        CMP AX, 0;                       has been exhausted.
        JNE SWTLP;
;
        POP AX;                         Restore work register and
        RET;                             return.
;
;
; ----- WAIT FOR REQUEST BIT.
;
;       This subroutine will monitor the Request bit in the
;       controller status port and wait for a certain length of
;       time for the Request bit to become set.
;
;       ENTRY:  No specific entry conditions.
;
;       EXIT:   The carry flag will be set to one if the Request
;                bit was set.
;               The carry flag will be cleared if a timeout
;                condition occurred while waiting for the Request
;                bit to be set.
;               The register AL has been destroyed.
;               WCNT has been destroyed.
;
;
IDWT:   PUSH DX;                        Save work register.
;
        MOV DX, SASI_STS;               Set status register port.
;
        MOV WCNT, 0;                    Set timeout count.
;
IDLP:   IN AL, DX;                      Check for DMA request bit
        RCR AL, 1;                       set, if so then exit.
        JNC IDWTC;
;
        POP DX;                         Restore work register and
        RET;                             return, if successful.
;
IDWTC:  DEC WCNT;                       Otherwise, decrement loop 
        JNZ IDLP;                        count and continue checking
;                                        until loop count is
;                                        exhausted.
;
        CLC;                            If timeout, clear the carry
;                                        flag.
;
        POP DX;                         Restore work register and
        RET;                             return.
;
HDDRV   ENDP
;
PROG    ENDS

; ----- MODULE REVISION HISTORY. ----------------------------------*
;
;       ORIGINAL: December 22, 1985  Mike Dougherty
;
;       REV 1.1:  February 14, 1986  Mike Dougherty
;       Repaired error where the two-drive code did not check for
;       an read/write/verify operation crossing drive boundries.
;       Also corrected timeout check for format commands,
;       controller select procedure, and incorrect error codes.
;
        END
